/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Use /dev/dsp as a channel, and the console to command it :).
 *
 * The full-duplex "simulation" is pretty weak.  This is generally a 
 * VERY BADLY WRITTEN DRIVER so please don't use it as a model for
 * writing a driver.
 * 
 * Copyright (C) 1999, Mark Spencer
 *
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 *
 * FreeBSD changes by Luigi Rizzo, 2005.04.18
 */

#include <asterisk/lock.h>
#include <asterisk/frame.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/module.h>
#include <asterisk/channel_pvt.h>
#include <asterisk/options.h>
#include <asterisk/pbx.h>
#include <asterisk/config.h>
#include <asterisk/cli.h>
#include <asterisk/utils.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef __linux
#include <linux/soundcard.h>
#elif defined(__FreeBSD__)
#include <sys/soundcard.h>
#else
#include <soundcard.h>
#endif
#include "busy.h"
#include "ringtone.h"
#include "ring10.h"
#include "answer.h"

/* Which device to use */
#if defined( __OpenBSD__ ) || defined( __NetBSD__ )
#define DEV_DSP "/dev/audio"
#else
#define DEV_DSP "/dev/dsp"
#endif

/* Lets use 160 sample frames, just like GSM.  */
/* this corresponds to 20ms of audio. */
#define FRAME_SIZE 160 // was 160

/*
 * When you set the frame size, you have to come up with
 * the right buffer format as well.
 * OSS lets you define a 'block' size (which should be a power of 2,
 * which power is specified in the lower 16 bits) and the number of
 * blocks allowed in the buffer (to avoid that the queue grows too large).
 * The latter is specified in the top 16 bits.
 * We use a block of 64 bytes (0x6), 5 blocks make a frame each sample
 * being 2 bytes, and we make room to store two buffers.
 * XXX the '10' is magic
 */

#define N_BLOCKS	(buffersize * 5 * 2)
/* 5 64-byte frames = one frame */
#define BUFFER_FMT (N_BLOCKS << 16) | (0x0006);

/* Don't switch between read/write modes faster than every 300 ms */
#define MIN_SWITCH_TIME 300


static int usecnt;


AST_MUTEX_DEFINE_STATIC(usecnt_lock);

static char *type = "Console";
static char *desc = "OSS Console Channel Driver";
static char *tdesc = "OSS Console Channel Driver";
static char *config = "oss.conf";

static char default_context[AST_MAX_EXTENSION] = "default";
static char language[MAX_LANGUAGE] = "";
static char oss_exten[AST_MAX_EXTENSION] = "s";



struct sound {
	int ind;
	char *desc;
	short *data;
	int datalen;
	int samplen;
	int silencelen;
	int repeat;
};

static struct sound sounds[] = {
	{ AST_CONTROL_RINGING, "RINGING", ringtone, sizeof(ringtone)/2, 16000, 32000, 1 },
	{ AST_CONTROL_BUSY, "BUSY", busy, sizeof(busy)/2, 4000, 4000, 1 },
	{ AST_CONTROL_CONGESTION, "CONGESTION", busy, sizeof(busy)/2, 2000, 2000, 1 },
	{ AST_CONTROL_RING, "RING10", ring10, sizeof(ring10)/2, 16000, 32000, 1 },
	{ AST_CONTROL_ANSWER, "ANSWER", answer, sizeof(answer)/2, 2200, 0, 0 },
	{ -1, NULL, 0, 0, 0, 0 },	/* end marker */
};



static struct chan_oss_pvt {
	/* We only have one OSS structure -- near sighted perhaps, but it
	   keeps this driver as simple as possible -- as it should be. */
    /*
     * cursound indicates which in struct sound we play. -1 means nothing,
     * any other value is a valid sound, in which case sampsent indicates
     * the next sample to send in [0..samplen + silencelen]
     * nosound is set to disable the audio data from the channel
     * (so we can play the tones etc.).
     */
    int sndcmd[2]; /* Sound command pipe */
    int cursound;	/* index of sound to send */
    int sampsent;	/* # of sound samples sent	*/
    int nosound;

    int total_blocks;	/* total blocks in the output device */
    int sounddev;
    enum { M_UNSET, M_FULL, M_READ, M_WRITE } duplex;
    int autoanswer;
    int autohangup;
    int hookstate;
    struct timeval lasttime;	/* last setformat */

    int silencesuppression;
    int silencethreshold;
	char device[64];	/* device to open */

    pthread_t sthread;

	struct ast_channel *owner;
	char exten[AST_MAX_EXTENSION];
	char context[AST_MAX_EXTENSION];
} oss = {
	.cursound = -1,
	.sounddev = -1,
	.duplex = M_UNSET, /* XXX check this */
	.autoanswer = 1,
	.autohangup = 1,
	.silencethreshold = 1000,
};

/*
 * returns true if too early to switch
 */
static int too_early(struct chan_oss_pvt *o)
{
	struct timeval tv;
	int ms;
	gettimeofday(&tv, NULL);
	ms = (tv.tv_sec - o->lasttime.tv_sec) * 1000 +
			(tv.tv_usec - o->lasttime.tv_usec) / 1000;
	if (ms < MIN_SWITCH_TIME)
		return -1;
	return 0;
}

/*
 * Returns the number of blocks used in the audio output channel
 */
static int
used_blocks(struct chan_oss_pvt *o)
{
    struct audio_buf_info info;
    static int warned=0;
    if (ioctl(o->sounddev, SNDCTL_DSP_GETOSPACE, &info)) {
	if (!warned) {
	    ast_log(LOG_WARNING, "Error reading output space\n");
	    warned++;
	}
	return 1;
    }
    if (o->total_blocks == 0) {
	ast_log(LOG_WARNING, "fragtotal %d size %d avail %d\n",
		info.fragstotal,
		info.fragsize,
		info.fragments);
	o->total_blocks = info.fragments;
    }
    return o->total_blocks - info.fragments;
}

static int soundcard_writeframe(struct chan_oss_pvt *o, short *data)
{	
	/* Write an exactly FRAME_SIZE sized of frame */
	int res;
	static int errors = 0;

	/*
	 * nothing spectacular.
	 * If the buffer is full just drop the extra, otherwise write
	 */
	res = used_blocks(o);
	if (res > 10) {	/* no room to write a block */
	    errors ++;
	    if (errors == 0)
	    	ast_log(LOG_WARNING, "write: used %d blocks (%d)\n", res, errors);
	    return 0;
	}
	errors = 0;
	res = write(o->sounddev, ((void *)data), FRAME_SIZE * 2);
	return res;
}

/*
 * handler for 'sound writable' events from the sound thread.
 * Builds a frame from the high level description of the sounds,
 * (tone+silence) and passes it to the audio device.
 */
static int send_sound(struct chan_oss_pvt *o)
{
	short myframe[FRAME_SIZE];
	int ofs = 0;
	int l_sampsent = o->sampsent;
	int l;
	struct sound *s;

	if (o->cursound < 0)	/* no sound to send */
		return 0;
	s = &sounds[o->cursound];
	/*
	 * prepare a frame
	 */
	
	for (ofs = 0; ofs < FRAME_SIZE; ofs += l) {
	    /* take chunks of sound and data until the buffer is full */
	    l = s->samplen - l_sampsent;	/* sound available */
	    if (l > 0) {
		if (l > FRAME_SIZE - ofs)
		    l = FRAME_SIZE - ofs;
		if (0)
		ast_log(LOG_WARNING, "send_sound sound %d/%d of %d into %d\n",
			l_sampsent, l, s->samplen, ofs);
		bcopy(s->data + l_sampsent, myframe + ofs, l*2);
		l_sampsent += l;
	    } else { /* no sound, maybe some silence */
		static short silence[FRAME_SIZE] = {0, };

	        l += s->silencelen;
		if (l > 0) {
		    if (l > FRAME_SIZE - ofs)
			l = FRAME_SIZE - ofs;
		    if (0)
		    ast_log(LOG_WARNING, "send_sound silence %d/%d of %d into %d\n",
			l_sampsent - s->samplen, l, s->silencelen, ofs);
		    bcopy(silence, myframe + ofs, l*2);
		    l_sampsent += l;
		} else { /* silence is over, restart sound if loop */
		    if (s->repeat == 0) {	/* last block */
			ast_log(LOG_WARNING, "send_sound last block\n");
			o->cursound = -1;
			o->nosound = 0;	/* allow audio data */
			if (ofs < FRAME_SIZE)	/* pad with silence */
			    bcopy(silence, myframe + ofs, (FRAME_SIZE - ofs)*2);
		    }
		    l_sampsent = 0;
		}
	    }
	}
	l = soundcard_writeframe(o, myframe);
	if (l > 0)
	    o->sampsent = l_sampsent;	/* update status */
	return 0;	/* fake success */
}

static void *sound_thread(void *unused)
{
	fd_set rfds;
	fd_set wfds;
	int max;
	int res;
	char ign[4096];
	if (read(oss.sounddev, ign, sizeof(ign)) < 0)
		ast_log(LOG_WARNING, "Read error on sound device: %s\n", strerror(errno));
	for(;;) {
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		max = oss.sndcmd[0];
		FD_SET(oss.sndcmd[0], &rfds);
		if (!oss.owner) {
			FD_SET(oss.sounddev, &rfds);
			if (oss.sounddev > max)
				max = oss.sounddev;
		}
		if (oss.cursound > -1) {
			FD_SET(oss.sounddev, &wfds);
			if (oss.sounddev > max)
				max = oss.sounddev;
		}
		res = ast_select(max + 1, &rfds, &wfds, NULL, NULL);
		if (res < 1) {
			ast_log(LOG_WARNING, "select failed: %s\n", strerror(errno));
			continue;
		}
		if (FD_ISSET(oss.sndcmd[0], &rfds)) { /* read which sound to play from the pipe */
			int i, what;

			read(oss.sndcmd[0], &what, sizeof(what));
			for (i = 0; sounds[i].ind != -1; i++)
				if (sounds[i].ind == what) {
					oss.cursound = i;
					oss.sampsent = 0;
					oss.nosound = 1; /* block other audio */
					ast_log(LOG_WARNING, "play %s\n", sounds[i].desc);
					break;
				}
			if (sounds[i].ind == -1)
				oss.cursound = -1;
			ast_log(LOG_WARNING, "cursound %d samplen %d silencelen %d\n",
				oss.cursound, oss.cursound >=0 ? sounds[oss.cursound].samplen : -1,
				oss.cursound >=0 ?  sounds[oss.cursound].silencelen : -1);
		}
		if (FD_ISSET(oss.sounddev, &rfds)) {
			/* Ignore read */
			if (read(oss.sounddev, ign, sizeof(ign)) < 0)
				ast_log(LOG_WARNING, "Read error on sound device: %s\n", strerror(errno));
		}
		if (FD_ISSET(oss.sounddev, &wfds))
			if (send_sound(&oss) < 0)
				ast_log(LOG_WARNING, "Failed to write sound\n");
	}
	/* Never reached */
	return NULL;
}

#if 0
static int calc_loudness(short *frame)
{
	int sum = 0;
	int x;
	for (x=0;x<FRAME_SIZE;x++) {
		if (frame[x] < 0)
			sum -= frame[x];
		else
			sum += frame[x];
	}
	sum = sum/FRAME_SIZE;
	return sum;
}

static int silence_suppress(short *buf)
{
#define SILBUF 3
	int loudness;
	static int silentframes = 0;
	static char silbuf[FRAME_SIZE * 2 * SILBUF];
	static int silbufcnt=0;
	if (!oss.silencesuppression)
		return 0;
	loudness = calc_loudness((short *)(buf));
	if (option_debug)
		ast_log(LOG_DEBUG, "loudness is %d\n", loudness);
	if (loudness < silencethreshold) {
		silentframes++;
		silbufcnt++;
		/* Keep track of the last few bits of silence so we can play
		   them as lead-in when the time is right */
		if (silbufcnt >= SILBUF) {
			/* Make way for more buffer */
			memmove(silbuf, silbuf + FRAME_SIZE * 2, FRAME_SIZE * 2 * (SILBUF - 1));
			silbufcnt--;
		}
		memcpy(silbuf + FRAME_SIZE * 2 * silbufcnt, buf, FRAME_SIZE * 2);
		if (silentframes > 10) {
			/* We've had plenty of silence, so compress it now */
			return 1;
		}
	} else {
		silentframes=0;
		/* Write any buffered silence we have, it may have something
		   important */
		if (silbufcnt) {
			write(oss.sounddev, silbuf, silbufcnt * FRAME_SIZE);
			silbufcnt = 0;
		}
	}
	return 0;
}
#endif

/*
 * reset and close the device if opened,
 * then open and initialize it in the desired mode,
 * trigger reads and writes so we can start using it.
 */
static int setformat(struct chan_oss_pvt *o, int mode)
{
	int fmt, desired, res, fd;
	static int warnedalready = 0;
	static int warnedalready2 = 0;

	if (o->sounddev >= 0) {
		ioctl(o->sounddev, SNDCTL_DSP_RESET, 0);
		close(o->sounddev);
		o->duplex = M_UNSET;
	}
	fd = o->sounddev = open(o->device, mode |O_NONBLOCK);
	if (o->sounddev < 0) {
		ast_log(LOG_WARNING, "Unable to re-open DSP device: %s\n",
			strerror(errno));
		return -1;
	}

	gettimeofday(&o->lasttime, NULL);
	fmt = AFMT_S16_LE;
	res = ioctl(fd, SNDCTL_DSP_SETFMT, &fmt);
	if (res < 0) {
		ast_log(LOG_WARNING, "Unable to set format to 16-bit signed\n");
		return -1;
	}
	switch (mode) {
	case O_RDWR:
	    res = ioctl(fd, SNDCTL_DSP_SETDUPLEX, 0);
	    /* Check to see if duplex set (FreeBSD Bug)*/
	    res = ioctl(fd, SNDCTL_DSP_GETCAPS, &fmt);
	    if (res == 0 && (fmt & DSP_CAP_DUPLEX)) {
		    if (option_verbose > 1) 
			ast_verbose(VERBOSE_PREFIX_2 "Console is full duplex\n");
		    o->duplex = M_FULL;
	    };
	    break;
	case O_WRONLY:
	    o->duplex = M_WRITE;
	    break;
	case O_RDONLY:
	    o->duplex = M_READ;
	    break;
	}

	fmt = 0;
	res = ioctl(fd, SNDCTL_DSP_STEREO, &fmt);
	if (res < 0) {
		ast_log(LOG_WARNING, "Failed to set audio device to mono\n");
		return -1;
	}
	/* 8000 Hz desired */
	desired = 8000;
	fmt = desired;
	res = ioctl(fd, SNDCTL_DSP_SPEED, &fmt);

	if (res < 0) {
		ast_log(LOG_WARNING, "Failed to set audio device to mono\n");
		return -1;
	}
	if (fmt != desired) {
		if (!warnedalready++)
			ast_log(LOG_WARNING, "Requested %d Hz, got %d Hz -- sound may be choppy\n", desired, fmt);
	}
#if 1
	/*
	 * on freebsd, SETFRAGMENT does not work very well on some cards.
	 * Better leave it out
	 */
	 
	// fmt = BUFFER_FMT;
	fmt = 0x8;	// 256-bytes fragment
	res = ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &fmt);
	if (res < 0) {
		if (!warnedalready2++)
			ast_log(LOG_WARNING, "Unable to set fragment size -- sound may be choppy\n");
	}
#endif
	/* XXX on some cards, we need SNDCTL_DSP_SETTRIGGER to start outputting */
	res = PCM_ENABLE_INPUT | PCM_ENABLE_OUTPUT;
	res = ioctl(fd, SNDCTL_DSP_SETTRIGGER, &res);
	/* it may fail if we are in half duplex, never mind */
	return 0;
}

/*
 * make sure output mode is available. Returns 0 if done,
 * 1 if too early to switch, -1 if error
 */
static int soundcard_setoutput(int force)
{
	if (oss.duplex == M_FULL || (oss.duplex == M_WRITE && !force))
		return 0;
	if (!force && too_early(&oss))
		return 1;
	if (setformat(&oss, O_WRONLY))
		return -1;
	return 0;
}

/*
 * make sure input mode is available. Returns 0 if done
 * 1 if too early to switch, -1 if error
 */
static int soundcard_setinput(int force)
{
	if (oss.duplex == M_FULL || (oss.duplex == M_READ && !force))
		return 0;
	if (!force && too_early(&oss))
		return 1;
	if (setformat(&oss, O_RDONLY))
		return -1;
	return 0;
}

static int oss_digit(struct ast_channel *c, char digit)
{
	ast_verbose( " << Console Received digit %c >> \n", digit);
	return 0;
}

static int oss_text(struct ast_channel *c, char *text)
{
	ast_verbose( " << Console Received text %s >> \n", text);
	return 0;
}

/* request to play a sound on the speaker */
#define	RING(x) { int what = x; write(oss.sndcmd[1], &what, sizeof(what)); }

static int oss_call(struct ast_channel *c, char *dest, int timeout)
{
	struct ast_frame f = { 0, };
	ast_verbose( " << Call placed to '%s' on console >> \n", dest);
	if (oss.autoanswer) {
		ast_verbose( " << Auto-answered >> \n" );
		f.frametype = AST_FRAME_CONTROL;
		f.subclass = AST_CONTROL_ANSWER;
		ast_queue_frame(c, &f);
	} else {
		ast_verbose( " << Type 'answer' to answer, or use 'autoanswer' for future calls >> \n");
		f.frametype = AST_FRAME_CONTROL;
		f.subclass = AST_CONTROL_RINGING;
		ast_queue_frame(c, &f);
		RING(AST_CONTROL_RING);
	}
	return 0;
}

static void answer_sound(void)
{
	RING(AST_CONTROL_ANSWER);
}

static int oss_answer(struct ast_channel *c)
{
	ast_verbose( " << Console call has been answered >> \n");
	answer_sound();	/* XXX do we really need it ? considering we shut down immediately... */
	ast_setstate(c, AST_STATE_UP);
	oss.cursound = -1;
	oss.nosound=0;
	return 0;
}

static int oss_hangup(struct ast_channel *c)
{
	oss.cursound = -1;
	c->pvt->pvt = NULL;
	oss.owner = NULL;
	ast_verbose( " << Hangup on console >> \n");
	ast_mutex_lock(&usecnt_lock);
	usecnt--;
	ast_mutex_unlock(&usecnt_lock);
	if (oss.hookstate) {
		if (oss.autoanswer || oss.autohangup) {
			/* Assume auto-hangup too */
			oss.hookstate = 0;
		} else {
			/* Make congestion noise */
			RING(AST_CONTROL_CONGESTION);
		}
	}
	return 0;
}

/* used for data coming from the network */
static int oss_write(struct ast_channel *chan, struct ast_frame *f)
{
	int res;
	int src;

	// ast_log(LOG_WARNING, "oss_write size %d\n", f->datalen);
	/* Immediately return if no sound is enabled */
	if (oss.nosound)
		return 0;
	/* Stop any currently playing sound */
	oss.cursound = -1;
	if (oss.duplex != M_FULL) {
		/* XXX check this, looks weird! */
		/* If we're half duplex, we have to switch to read mode
		   to honor immediate needs if necessary */
		res = soundcard_setinput(1); /* force set if not full_duplex */
		if (res < 0) {
			ast_log(LOG_WARNING, "Unable to set device to input mode\n");
			return -1;
		}
		return 0;
	}
	res = soundcard_setoutput(0);
	if (res < 0) {
		ast_log(LOG_WARNING, "Unable to set output device\n");
		return -1;
	} else if (res > 0) {
		/* The device is still in read mode, and it's too soon to change it,
		   so just pretend we wrote it */
		return 0;
	}
	/*
	 * we could receive a sample which is not a multiple of our FRAME_SIZE,
	 * so we buffer it locally and write to the device in FRAME_SIZE
	 * chunks, keeping the residue stored for future use.
	 */

	src = 0; /* read position into f->data */
	while ( src < f->datalen ) {
	    static char buf[FRAME_SIZE*2];
	    static int dst = 0;
	    int l = sizeof(buf) - dst; /* how much room in the buffer */

	    if (f->datalen - src >= l) {	/* enough to fill a frame */
		memcpy(buf + dst, f->data + src, l);
		soundcard_writeframe(&oss, (short *)buf);
		src += l;
		dst = 0;
	    } else { /* copy residue */
		l = f->datalen - src;
		memcpy(buf + dst, f->data + src, l);
		src += l;	/* but really, we are done */
		dst += l;
	    }
	}
	return 0;
}

static struct ast_frame *oss_read(struct ast_channel *chan)
{
	static struct ast_frame f;
	static char buf[FRAME_SIZE * 2 + AST_FRIENDLY_OFFSET];
	static int readpos = 0;
	int res;
	
#if 0
	ast_log(LOG_DEBUG, "oss_read()\n");
#endif
		
	f.frametype = AST_FRAME_NULL;
	f.subclass = 0;
	f.samples = 0;
	f.datalen = 0;
	f.data = NULL;
	f.offset = 0;
	f.src = type;
	f.mallocd = 0;
	f.delivery.tv_sec = 0;
	f.delivery.tv_usec = 0;
	
	res = soundcard_setinput(0);
	if (res < 0) {
		ast_log(LOG_WARNING, "Unable to set input mode\n");
		return NULL;
	}
	if (res > 0) {	/* too early to switch ? */
		/* Theoretically shouldn't happen, but anyway, return a NULL frame */
		return &f;
	}

	res = read(oss.sounddev, buf + AST_FRIENDLY_OFFSET + readpos, FRAME_SIZE * 2 - readpos);
	// ast_log(LOG_WARNING, "oss_read() fd %d got %d\n", oss.sounddev, res);
	if (res < 0) /* audio data not ready, return a NULL frame */
		return &f;
	readpos += res;
	
	if (readpos >= FRAME_SIZE * 2) {
		/* A real frame */
		readpos = 0;
		if (chan->_state != AST_STATE_UP) {
			/* Don't transmit unless it's up */
			return &f;
		}
		f.frametype = AST_FRAME_VOICE;
		f.subclass = AST_FORMAT_SLINEAR;
		f.samples = FRAME_SIZE;
		f.datalen = FRAME_SIZE * 2;
		f.data = buf + AST_FRIENDLY_OFFSET;
		f.offset = AST_FRIENDLY_OFFSET;
		f.src = type;
		f.mallocd = 0;
		f.delivery.tv_sec = 0;
		f.delivery.tv_usec = 0;
#if 0
		{ static int fd = -1;
		  if (fd < 0)
		  	fd = open("output.raw", O_RDWR | O_TRUNC | O_CREAT);
		  write(fd, f.data, f.datalen);
		}
#endif		
	}
	return &f;
}

static int oss_fixup(struct ast_channel *oldchan, struct ast_channel *newchan)
{
	struct chan_oss_pvt *p = newchan->pvt->pvt;
	p->owner = newchan;
	return 0;
}

static int oss_indicate(struct ast_channel *chan, int cond)
{
	int res;
	switch(cond) {
	case AST_CONTROL_BUSY:
	case AST_CONTROL_CONGESTION:
	case AST_CONTROL_RINGING:
		res = cond;
		break;
	case -1:
		oss.cursound = -1;
		return 0;
	default:
		ast_log(LOG_WARNING, "Don't know how to display condition %d on %s\n", cond, chan->name);
		return -1;
	}
	if (res > -1) {
		RING(res);
	}
	return 0;	
}

static struct ast_channel *oss_new(struct chan_oss_pvt *oss, int state)
{
	struct ast_channel *tmp;
	struct ast_channel_pvt *pvt;

	tmp = ast_channel_alloc(1);
	if (tmp == NULL)
		return NULL;
	snprintf(tmp->name, sizeof(tmp->name), "OSS/%s", oss->device + 5);
	tmp->type = type;
	tmp->fds[0] = oss->sounddev;
	tmp->nativeformats = AST_FORMAT_SLINEAR;
	pvt = tmp->pvt;
	pvt->pvt = oss;
#if 1
	pvt->send_digit = oss_digit;
	pvt->send_text = oss_text;
	pvt->hangup = oss_hangup;
	pvt->answer = oss_answer;
	pvt->read = oss_read;
	pvt->call = oss_call;
	pvt->write = oss_write;
	pvt->indicate = oss_indicate;
	pvt->fixup = oss_fixup;
#endif
	if (strlen(oss->context))
		strncpy(tmp->context, oss->context, sizeof(tmp->context)-1);
	if (strlen(oss->exten))
		strncpy(tmp->exten, oss->exten, sizeof(tmp->exten)-1);
	if (strlen(language))
		strncpy(tmp->language, language, sizeof(tmp->language)-1);
	oss->owner = tmp;
	ast_setstate(tmp, state);
	ast_mutex_lock(&usecnt_lock);
	usecnt++;
	ast_mutex_unlock(&usecnt_lock);
	ast_update_use_count();
	if (state != AST_STATE_DOWN) {
		if (ast_pbx_start(tmp)) {
			ast_log(LOG_WARNING, "Unable to start PBX on %s\n", tmp->name);
			ast_hangup(tmp);
			tmp = NULL;
			/* XXX what about oss->owner and the channel itself ? */
		}
	}
	return tmp;
}

static struct ast_channel *oss_request(char *type, int format, void *data)
{
	int oldformat = format;
	struct ast_channel *tmp;
	format &= AST_FORMAT_SLINEAR;
	if (!format) {
		ast_log(LOG_NOTICE, "Asked to get a channel of format '%d'\n", oldformat);
		return NULL;
	}
	if (oss.owner) {
		ast_log(LOG_NOTICE, "Already have a call on the OSS channel\n");
		return NULL;
	}
	tmp= oss_new(&oss, AST_STATE_DOWN);
	if (!tmp) {
		ast_log(LOG_WARNING, "Unable to create new OSS channel\n");
	}
	return tmp;
}

static int console_autoanswer(int fd, int argc, char *argv[])
{
	if ((argc != 1) && (argc != 2))
		return RESULT_SHOWUSAGE;
	if (argc == 1) {
		ast_cli(fd, "Auto answer is %s.\n", oss.autoanswer ? "on" : "off");
		return RESULT_SUCCESS;
	} else {
		if (!strcasecmp(argv[1], "on"))
			oss.autoanswer = -1;
		else if (!strcasecmp(argv[1], "off"))
			oss.autoanswer = 0;
		else
			return RESULT_SHOWUSAGE;
	}
	return RESULT_SUCCESS;
}

static char *autoanswer_complete(char *line, char *word, int pos, int state)
{
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif
	int l = strlen(word);

	switch(state) {
	case 0:
		if (l && !strncasecmp(word, "on", MIN(l, 2)))
			return strdup("on");
	case 1:
		if (l && !strncasecmp(word, "off", MIN(l, 3)))
			return strdup("off");
	default:
		return NULL;
	}
	return NULL;
}

static char autoanswer_usage[] =
"Usage: autoanswer [on|off]\n"
"       Enables or disables autoanswer feature.  If used without\n"
"       argument, displays the current on/off status of autoanswer.\n"
"       The default value of autoanswer is in 'oss.conf'.\n";

static int console_answer(int fd, int argc, char *argv[])
{
	struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_ANSWER };
	if (argc != 1)
		return RESULT_SHOWUSAGE;
	if (!oss.owner) {
		ast_cli(fd, "No one is calling us\n");
		return RESULT_FAILURE;
	}
	oss.hookstate = 1;
	oss.cursound = -1;
	ast_queue_frame(oss.owner, &f);
	answer_sound();
	return RESULT_SUCCESS;
}

static char sendtext_usage[] =
"Usage: send text <message>\n"
"       Sends a text message for display on the remote terminal.\n";

static int console_sendtext(int fd, int argc, char *argv[])
{
	int tmparg = 2;
	char text2send[256] = "";
	struct ast_frame f = { 0, };
	if (argc < 2)
		return RESULT_SHOWUSAGE;
	if (!oss.owner) {
		ast_cli(fd, "No one is calling us\n");
		return RESULT_FAILURE;
	}
	if (strlen(text2send))
		ast_cli(fd, "Warning: message already waiting to be sent, overwriting\n");
	text2send[0] = '\0';
	while(tmparg < argc) {
		strncat(text2send, argv[tmparg++], sizeof(text2send) - strlen(text2send) - 1);
		strncat(text2send, " ", sizeof(text2send) - strlen(text2send) - 1);
	}
	if (strlen(text2send)) {
		f.frametype = AST_FRAME_TEXT;
		f.subclass = 0;
		f.data = text2send;
		f.datalen = strlen(text2send);
		ast_queue_frame(oss.owner, &f);
	}
	return RESULT_SUCCESS;
}

static char answer_usage[] =
"Usage: answer\n"
"       Answers an incoming call on the console (OSS) channel.\n";

static int console_hangup(int fd, int argc, char *argv[])
{
	if (argc != 1)
		return RESULT_SHOWUSAGE;
	oss.cursound = -1;
	if (!oss.owner && !oss.hookstate) {
		ast_cli(fd, "No call to hangup up\n");
		return RESULT_FAILURE;
	}
	oss.hookstate = 0;
	if (oss.owner) {
		ast_queue_hangup(oss.owner);
	}
	return RESULT_SUCCESS;
}

static char hangup_usage[] =
"Usage: hangup\n"
"       Hangs up any call currently placed on the console.\n";


static int console_dial(int fd, int argc, char *argv[])
{
	char tmp[256], *tmp2;
	char *mye, *myc;
	int x;
	struct ast_frame f = { AST_FRAME_DTMF, 0 };
	if ((argc != 1) && (argc != 2))
		return RESULT_SHOWUSAGE;
	if (oss.owner) {
		if (argc == 2) {
			for (x=0;x<strlen(argv[1]);x++) {
				f.subclass = argv[1][x];
				ast_queue_frame(oss.owner, &f);
			}
		} else {
			ast_cli(fd, "You're already in a call.  You can use this only to dial digits until you hangup\n");
			return RESULT_FAILURE;
		}
		return RESULT_SUCCESS;
	}
	mye = oss_exten;
	myc = default_context;
	if (argc == 2) {
		char *stringp=NULL;
		strncpy(tmp, argv[1], sizeof(tmp)-1);
		stringp=tmp;
		strsep(&stringp, "@");
		tmp2 = strsep(&stringp, "@");
		if (strlen(tmp))
			mye = tmp;
		if (tmp2 && strlen(tmp2))
			myc = tmp2;
	}
	if (ast_exists_extension(NULL, myc, mye, 1, NULL)) {
		strncpy(oss.exten, mye, sizeof(oss.exten)-1);
		strncpy(oss.context, myc, sizeof(oss.context)-1);
		oss.hookstate = 1;
		oss_new(&oss, AST_STATE_RINGING);
	} else
		ast_cli(fd, "No such extension '%s' in context '%s'\n", mye, myc);
	return RESULT_SUCCESS;
}

static char dial_usage[] =
"Usage: dial [extension[@context]]\n"
"       Dials a given extensison (and context if specified)\n";

static int console_transfer(int fd, int argc, char *argv[])
{
	char tmp[256];
	char *context;
	if (argc != 2)
		return RESULT_SHOWUSAGE;
	if (oss.owner && oss.owner->bridge) {
		strncpy(tmp, argv[1], sizeof(tmp) - 1);
		context = strchr(tmp, '@');
		if (context) {
			*context = '\0';
			context++;
		} else
			context = oss.owner->context;
		if (ast_exists_extension(oss.owner->bridge, context, tmp, 1, oss.owner->bridge->callerid)) {
			ast_cli(fd, "Whee, transferring %s to %s@%s.\n", 
					oss.owner->bridge->name, tmp, context);
			if (ast_async_goto(oss.owner->bridge, context, tmp, 1))
				ast_cli(fd, "Failed to transfer :(\n");
		} else {
			ast_cli(fd, "No such extension exists\n");
		}
	} else {
		ast_cli(fd, "There is no call to transfer\n");
	}
	return RESULT_SUCCESS;
}

static char transfer_usage[] =
"Usage: transfer <extension>[@context]\n"
"       Transfers the currently connected call to the given extension (and\n"
"context if specified)\n";

static struct ast_cli_entry myclis[] = {
	{ { "answer", NULL }, console_answer, "Answer an incoming console call", answer_usage },
	{ { "hangup", NULL }, console_hangup, "Hangup a call on the console", hangup_usage },
	{ { "dial", NULL }, console_dial, "Dial an extension on the console", dial_usage },
	{ { "transfer", NULL }, console_transfer, "Transfer a call to a different extension", transfer_usage },
	{ { "send", "text", NULL }, console_sendtext, "Send text to the remote device", sendtext_usage },
	{ { "autoanswer", NULL }, console_autoanswer, "Sets/displays autoanswer", autoanswer_usage, autoanswer_complete }
};

int load_module()
{
	int res;
	int x;
	struct ast_config *cfg;

	res = pipe(oss.sndcmd);
	if (res) {
		ast_log(LOG_ERROR, "Unable to create pipe\n");
		return -1;
	}
	/* load config file */
	if ((cfg = ast_load(config))) {
		struct ast_variable *v = ast_variable_browse(cfg, "general");
		while(v) {
			if (!strcasecmp(v->name, "autoanswer"))
				oss.autoanswer = ast_true(v->value);
			else if (!strcasecmp(v->name, "autohangup"))
				oss.autohangup = ast_true(v->value);
			else if (!strcasecmp(v->name, "oss.silencesuppression"))
				oss.silencesuppression = ast_true(v->value);
			else if (!strcasecmp(v->name, "silencethreshold"))
				oss.silencethreshold = atoi(v->value);
			else if (!strcasecmp(v->name, "device"))
				strncpy(oss.device, v->value, sizeof(oss.device)-1);
			else if (!strcasecmp(v->name, "context"))
				strncpy(default_context, v->value, sizeof(default_context)-1);
			else if (!strcasecmp(v->name, "language"))
				strncpy(language, v->value, sizeof(language)-1);
			else if (!strcasecmp(v->name, "extension"))
				strncpy(oss_exten, v->value, sizeof(oss_exten)-1);
			v=v->next;
		}
		ast_destroy(cfg);
	}
	if (!strlen(oss.device))
		strncpy(oss.device, DEV_DSP, sizeof(oss.device)-1);
	if (setformat(&oss, O_RDWR) < 0) {	/* open device */
		if (option_verbose > 1) {
			ast_verbose(VERBOSE_PREFIX_2 "No sound card detected -- console channel will be unavailable\n");
			ast_verbose(VERBOSE_PREFIX_2 "Turn off OSS support by adding 'noload=chan_oss.so' in /etc/asterisk/modules.conf\n");
		}
		return 0;
	}
	soundcard_setinput(1); /* force set if not full_duplex */
	if (oss.duplex != M_FULL)
		ast_log(LOG_WARNING, "XXX I don't work right with non-full duplex sound cards XXX\n");
	res = ast_channel_register(type, tdesc, AST_FORMAT_SLINEAR, oss_request);
	if (res < 0) {
		ast_log(LOG_ERROR, "Unable to register channel class '%s'\n", type);
		return -1;
	}
	for (x=0;x<sizeof(myclis)/sizeof(struct ast_cli_entry); x++)
		ast_cli_register(myclis + x);
	ast_pthread_create(&oss.sthread, NULL, sound_thread, NULL);
	return 0;
}



int unload_module()
{
	int x;
	for (x=0;x<sizeof(myclis)/sizeof(struct ast_cli_entry); x++)
		ast_cli_unregister(myclis + x);
	close(oss.sounddev);
	if (oss.sndcmd[0] > 0) {
		close(oss.sndcmd[0]);
		close(oss.sndcmd[1]);
	}
	if (oss.owner)
		ast_softhangup(oss.owner, AST_SOFTHANGUP_APPUNLOAD);
	if (oss.owner)
		return -1;
	/* XXX what about the thread ? */
	return 0;
}

char *description()
{
	return desc;
}

int usecount()
{
	int res;
	ast_mutex_lock(&usecnt_lock);
	res = usecnt;
	ast_mutex_unlock(&usecnt_lock);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
